from pathlib import Path

import numpy as np


def dump_all_yield_stresses(VE_response, yield_point, dump_path):
    all_VE_responses = VE_response
    yield_stresses = []
    for VE_resp_i in all_VE_responses:
        strain = VE_resp_i["phase_data"]["vol_avg_equivalent_plastic_strain"]["data"][:]
        yield_idx = np.argmin(np.abs(strain - yield_point))
        stress_i = VE_resp_i["phase_data"]["vol_avg_equivalent_stress"]["data"][yield_idx]
        yield_stresses.append(stress_i)

    with Path(dump_path).open("wt") as fh:
        for i in yield_stresses:
            fh.write(f"{i:.5f}\n")
    return {}



from matflow.param_classes.load import LoadCase
from matflow.param_classes.orientations import Orientations
from matflow.param_classes.seeds import MicrostructureSeeds

__all__ = [
    "LoadCase",
    "Orientations",
    "MicrostructureSeeds",
]



from damask_parse.readers import read_spectral_stdout


def read_log(damask_stdout_file):
    try:
        return read_spectral_stdout(path=damask_stdout_file, encoding="utf8")
    except UnicodeDecodeError:
        # Docker on Windows...
        return read_spectral_stdout(path=damask_stdout_file, encoding="utf16")



from __future__ import annotations

import logging

import numpy as np
from hpcflow.sdk.core.parameters import ParameterValue

logger = logging.getLogger(__name__)


class LoadCase(ParameterValue):

    _typ = "load_case"

    def __init__(
        self,
        direction: str,
        total_time: float,
        num_increments: int,
        def_grad_rate: np.typing.ArrayLike,
        def_grad_aim: np.typing.ArrayLike,
        stress: np.typing.ArrayLike,
        dump_frequency: int,
    ) -> None:
        self.direction = direction
        self.total_time = total_time
        self.num_increments = num_increments
        self.def_grad_rate = def_grad_rate
        self.def_grad_aim = def_grad_aim
        self.direction = direction
        self.stress = stress
        self.dump_frequency = dump_frequency

    @staticmethod
    def helper_method(a):
        pass

    @classmethod
    def uniaxial(
        cls,
        total_time: float,
        num_increments: int,
        direction: str,
        target_strain_rate: float = None,
        target_strain: float = None,
        dump_frequency: int = 1,
    ) -> LoadCase:
        # Validation:
        msg = "Specify either `target_strain_rate` or `target_strain`."
        if all([t is None for t in [target_strain_rate, target_strain]]):
            raise ValueError(msg)
        if all([t is not None for t in [target_strain_rate, target_strain]]):
            raise ValueError(msg)

        if target_strain_rate is not None:
            def_grad_val = target_strain_rate
        else:
            def_grad_val = target_strain

        dir_idx = ["x", "y", "z"]
        try:
            loading_dir_idx = dir_idx.index(direction)
        except ValueError:
            msg = (
                f'Loading direction "{direction}" not allowed. It should be one of "x", '
                f'"y" or "z".'
            )
            raise ValueError(msg)

        dg_arr = np.ma.masked_array(np.zeros((3, 3)), mask=np.eye(3))
        stress_arr = np.ma.masked_array(np.zeros((3, 3)), mask=np.logical_not(np.eye(3)))

        dg_arr[loading_dir_idx, loading_dir_idx] = def_grad_val
        dg_arr.mask[loading_dir_idx, loading_dir_idx] = False
        stress_arr.mask[loading_dir_idx, loading_dir_idx] = True

        def_grad_aim = dg_arr if target_strain is not None else None
        def_grad_rate = dg_arr if target_strain_rate is not None else None

        load_case = {
            "direction": direction,
            "total_time": total_time,
            "num_increments": num_increments,
            "def_grad_rate": def_grad_rate,
            "def_grad_aim": def_grad_aim,
            "stress": stress_arr,
            "dump_frequency": dump_frequency,
        }
        return cls(**load_case)



import copy
import numpy as np
from damask import Grid
from damask_parse.utils import validate_volume_element, validate_orientations


def generate_volume_element_voronoi(
    microstructure_seeds,
    VE_grid_size,
    homog_label,
    orientations,
):
    grid_obj = Grid.from_Voronoi_tessellation(
        cells=np.array(VE_grid_size),
        size=np.array(microstructure_seeds.box_size),
        seeds=np.array(microstructure_seeds.position),
        periodic=True,
    )
    # TODO: this needs some more thought.

    if orientations is None:
        oris = copy.deepcopy(microstructure_seeds.orientations)

    else:
        # see `LatticeDirection` enum:
        align_lookup = {
            "A": "a",
            "B": "b",
            "C": "c",
            "A_STAR": "a*",
            "B_STAR": "b*",
            "C_STAR": "c*",
        }
        unit_cell_alignment = {
            "x": align_lookup[orientations.unit_cell_alignment.x.name],
            "y": align_lookup[orientations.unit_cell_alignment.y.name],
            "z": align_lookup[orientations.unit_cell_alignment.z.name],
        }
        type_lookup = {
            "QUATERNION": "quat",
            "EULER": "euler",
        }
        type_ = type_lookup[orientations.representation.type.name]
        oris = {
            "type": type_,
            "unit_cell_alignment": unit_cell_alignment,
        }

        if type_ == "quat":
            quat_order = orientations.representation.quat_order.name.lower().replace(
                "_", "-"
            )
            oris["quaternions"] = np.array(orientations.data)
            oris["quat_component_ordering"] = quat_order
        elif type_ == "euler":
            oris["euler_angles"] = np.array(orientations.data)
            oris["euler_degrees"] = orientations.representation.euler_is_degrees

    oris = validate_orientations(oris)

    num_grains = len(microstructure_seeds.position)
    const_phase_lab = np.array([microstructure_seeds.phase_label])[
        np.zeros(num_grains, dtype=int)
    ]

    ori_idx = np.arange(num_grains)

    volume_element = {
        "size": grid_obj.size.astype(float).tolist(),
        "grid_size": grid_obj.cells.tolist(),
        "orientations": oris,
        "element_material_idx": grid_obj.material,
        "constituent_material_idx": np.arange(num_grains),
        "constituent_material_fraction": np.ones(num_grains),
        "constituent_phase_label": const_phase_lab,
        "constituent_orientation_idx": ori_idx,
        "material_homog": np.full(num_grains, homog_label),
    }
    volume_element = validate_volume_element(volume_element)
    return {"volume_element": volume_element}



from pathlib import Path
from typing import Dict, List, Union

from damask_parse.utils import generate_viz


def viz_damask_HDF5(
    damask_hdf5_file: Path,
    damask_viz: Union[Dict, List],
    VE_response: Dict,
):
    generate_viz(hdf5_path=damask_hdf5_file, viz_spec=damask_viz, parsed_outs=VE_response)



from dataclasses import dataclass
import enum
from pathlib import Path
from typing import Optional

import numpy as np
import zarr

from hpcflow.sdk.core.parameters import ParameterValue


def get_enum_by_name_or_val(cls, key):
    """Retrieve an enum by name or value, assuming uppercase names and integer values."""
    if key is None or isinstance(key, cls):
        return key
    elif isinstance(key, (int, float)):
        return cls(int(key))  # retrieve by value
    else:
        return getattr(cls, key.upper())  # retrieve by name


class EulerDefinition(enum.Enum):
    def __new__(cls, value, rotation_order, doc=None):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.rotation_order = rotation_order
        obj.__doc__ = doc
        return obj

    BUNGE = (0, "ZXZ", "Convention typically used in crystallography.")


class QuatOrder(enum.Enum):
    """Order in which the four quaternion components listed.

    Dream3D [1] uses vector-scalar ordering, whereas most other programs seem to use
    scalar-vector ordering.

    References
    ----------
    [1] http://dream3d.bluequartz.net/Help/Filters/OrientationAnalysisFilters/ConvertQuaternion/

    """

    SCALAR_VECTOR = 0
    VECTOR_SCALAR = 1


class OrientationRepresentationType(enum.Enum):

    QUATERNION = 0
    EULER = 1


@dataclass
class OrientationRepresentation:

    type: OrientationRepresentationType
    euler_definition: Optional[EulerDefinition] = None
    euler_is_degrees: Optional[bool] = None
    quat_order: Optional[QuatOrder] = None

    def __post_init__(self):
        self.type = get_enum_by_name_or_val(OrientationRepresentationType, self.type)
        if self.type is OrientationRepresentationType.EULER:
            if self.euler_definition is None:
                raise ValueError("Must specify `euler_definition`.")
            if self.euler_is_degrees is None:
                raise ValueError("Must specify `euler_is_degrees`.")
        elif self.type is OrientationRepresentationType.QUATERNION:
            if self.quat_order is None:
                raise ValueError("Must specify `quat_order`.")
        self.euler_definition = get_enum_by_name_or_val(
            EulerDefinition, self.euler_definition
        )
        self.quat_order = get_enum_by_name_or_val(QuatOrder, self.quat_order)


class LatticeDirection(enum.Enum):
    # real-space directions:
    A = 0
    B = 1
    C = 2

    # reciprocal-space directions
    A_STAR = 3
    B_STAR = 4
    C_STAR = 5


@dataclass
class UnitCellAlignment(ParameterValue):

    _typ = "unit_cell_alignment"

    x: Optional[LatticeDirection] = None
    y: Optional[LatticeDirection] = None
    z: Optional[LatticeDirection] = None

    def __post_init__(self):
        self.x = get_enum_by_name_or_val(LatticeDirection, self.x)
        self.y = get_enum_by_name_or_val(LatticeDirection, self.y)
        self.z = get_enum_by_name_or_val(LatticeDirection, self.z)

    @classmethod
    def from_hex_convention_DAMASK(cls):
        # TODO: check!
        return cls(x=LatticeDirection.A, y=LatticeDirection.B_STAR, z=LatticeDirection.C)

    @classmethod
    def from_hex_convention_MTEX(cls):
        """Generate a unit cell alignment from MTEX's default convention for hexagonal
        symmetry.

        Tested using this command in MTEX: `crystalSymmetry("hexagonal").alignment`

        """
        return cls(
            x=LatticeDirection.A_STAR,
            y=LatticeDirection.B,
            z=LatticeDirection.C_STAR,
        )


@dataclass
class Orientations(ParameterValue):

    _typ = "orientations"

    data: np.ndarray
    unit_cell_alignment: UnitCellAlignment
    representation: OrientationRepresentation

    def __post_init__(self):
        if not isinstance(self.representation, OrientationRepresentation):
            self.representation = OrientationRepresentation(**self.representation)

        if not isinstance(self.unit_cell_alignment, UnitCellAlignment):
            self.unit_cell_alignment = UnitCellAlignment(**self.unit_cell_alignment)

    @classmethod
    def save_from_HDF5_group(cls, group, param_id: int, workflow):
        """Save orientation data from an HDF5 group to a persistent workflow.

        We avoid loading the data into memory all at once by firstly generating an
        `Orientations` object with a small data array, and then copying from the HDF5
        group directly into the newly created Zarr group.

        """

        obj = cls(
            data=np.array([0]),
            representation=group.attrs.get("representation"),
            unit_cell_alignment=dict(
                zip(("x", "y", "z"), group.attrs.get("unit_cell_alignment"))
            ),
        )
        workflow.set_parameter_value(param_id=param_id, value=obj, commit=True)

        # now replace placeholder data with correct data:
        zarr_grp, dataset_name = workflow._store._get_array_group_and_dataset(
            mode="r+",
            param_id=param_id,
            data_path=["data"],
        )
        zarr.copy(
            source=group["data"],
            dest=zarr_grp,
            name=dataset_name,
            if_exists="replace",
        )

    def dump_to_HDF5_group(self, HDF5_group):
        pass

    @classmethod
    def from_JSON_like(cls, data, ori_format):
        """For custom initialisation via YAML or JSON."""
        return cls(data=np.asarray(data), ori_format=ori_format)

    # def to_json_like(self, dct=None, shared_data=None, exclude=None, _depth=0):
    #     return

    @classmethod
    def from_random(cls, number):
        return cls(
            data=cls.quat_sample_random(number),
            unit_cell_alignment=UnitCellAlignment.from_hex_convention_DAMASK(),
            representation=OrientationRepresentation(
                type=OrientationRepresentationType.QUATERNION,
                quat_order=QuatOrder.SCALAR_VECTOR,
            ),
        )

    @classmethod
    def from_file(
        cls,
        path,
        representation,
        unit_cell_alignment,
        number=None,
        start_index=0,
        delimiter=" ",
    ):
        rep = OrientationRepresentation(**representation)
        data = []
        with Path(path).open("rt") as fh:
            for idx, line in enumerate(fh):
                line = line.strip()
                if not line or idx < start_index:
                    continue
                elif len(data) < number if number is not None else True:
                    data.append([float(i) for i in line.split(delimiter)])
        if number is not None and len(data) < number:
            raise ValueError("Not enough orientations in the file.")

        return cls(
            data=np.asarray(data),
            representation=rep,
            unit_cell_alignment=unit_cell_alignment,
        )

    @staticmethod
    def quat_sample_random(number):
        """Generate random uniformly distributed unit quaternions.

        Parameters
        ----------
        number : int
            How many quaternions to generate.

        Returns
        -------
        quats : ndarray of shape (number, 4)

        References
        ----------
        https://stackoverflow.com/a/44031492/5042280
        http://planning.cs.uiuc.edu/node198.html
        """

        rand_nums = np.random.random((number, 3))
        quats = np.array(
            [
                np.sqrt(1 - rand_nums[:, 0]) * np.sin(2 * np.pi * rand_nums[:, 1]),
                np.sqrt(1 - rand_nums[:, 0]) * np.cos(2 * np.pi * rand_nums[:, 1]),
                np.sqrt(rand_nums[:, 0]) * np.sin(2 * np.pi * rand_nums[:, 2]),
                np.sqrt(rand_nums[:, 0]) * np.cos(2 * np.pi * rand_nums[:, 2]),
            ]
        ).T

        return quats



from PyInstaller.utils.hooks import collect_data_files

hiddenimports = ["matflow.data", "matflow.param_classes"]
datas = collect_data_files("matflow.data")
datas += collect_data_files(
    "matflow.tests",
    include_py_files=True,
    excludes=("**/__pycache__",),
)
datas += collect_data_files(
    "matflow.data.scripts",
    include_py_files=True,
    excludes=("**/__pycache__",),
)



from damask import seeds
from damask import Rotation
import numpy as np


def seeds_from_random(
    VE_size,
    num_grains,
    phase_label,
    orientation_coordinate_system=None,
):

    VE_size = np.array(VE_size)
    position = seeds.from_random(VE_size, num_grains)
    rotation = Rotation.from_random(shape=(num_grains,))
    out = {
        "microstructure_seeds": {
            "position": position,
            "orientations": {
                "type": "quat",
                "quaternions": rotation.quaternion,
                "quat_component_ordering": "scalar-vector",
                "orientation_coordinate_system": orientation_coordinate_system,
                "unit_cell_alignment": {
                    "x": "a",
                    "z": "c",
                },
                "P": -1,
            },
            "size": VE_size,
            "random_seed": None,
            "phase_labels": [phase_label],
        }
    }
    return out
